__all__ = [
    'BaseLoader',
]

from alviss.structs import *
from .interface import *
from ccptools.tpu import iters
from ccptools.tpu import strimp
import os
import logging
log = logging.getLogger(__file__)


_VAR_PATTERN = re.compile(r'\$\{(?P<var>(?:[^=}\s]+)(?:!?=[^}]*)?)}')


class BaseLoader(IAlvissLoader, abc.ABC):
    def __init__(self):
        self._file_name = ''
        self._file_path = ''
        self._data: Dict = {}

        self._unresolved: Dict[Sequence[str], str] = {}
        self._extends: List[Sequence[str, Sequence[str]]] = []
        self._include: List[Sequence[str, Sequence[str]]] = []
        self._chains: List[Tuple[Sequence[str], str]] = []

        self._resolve_count = 0
        self._special_key_map: Dict[str, Callable] = {
            '__extends__': self._special_key_extends,
            '__extend__': self._special_key_extends,
            '__include__': self._special_key_include,
            '__includes__': self._special_key_include,
            '__fidelius__': self._special_key_fidelius,
        }

        self._skip_resolve: bool = False
        self._skip_env_loading: bool = False
        self._skip_includes: bool = False
        self._skip_extends: bool = False
        self._skip_fidelius: bool = False
        self._skip_py_inject: bool = False

        self._fidelius_keys: Dict[Sequence[str], str] = {}
        self._fidelius_mode: FideliusMode = FideliusMode.ON_DEMAND
        self._fidelius_kwargs: Optional[Dict[str, Any]] = None
        self._find_fidelius_mode()

    def set_fidelius_mode(self, mode: Union[FideliusMode, str, int]):
        self._fidelius_mode = FideliusMode.from_any(mode)

    def get_fidelius_mode(self) -> FideliusMode:
        return self._fidelius_mode

    def _find_fidelius_mode(self):
        env = os.environ.get('ALVISS_FIDELIUS_MODE', FideliusMode.ON_DEMAND)
        if env == '':
            env = FideliusMode.ON_DEMAND
        self.set_fidelius_mode(env)
        if self.get_fidelius_mode() == FideliusMode.ENABLED:
            try:
                import fidelius
            except ImportError:
                log.error('ALVISS_FIDELIUS_MODE is ENABLED but fidelius is not installed')
                raise AlvissFideliusNotInstalledError('ALVISS_FIDELIUS_MODE is ENABLED but fidelius is not installed')

    @property
    def data(self) -> Dict:
        return self._data

    def load_file(self,
                  file_name: str,
                  no_resolve: bool = False,
                  no_extend: bool = False,
                  no_includes: bool = False,
                  no_env_load: bool = False,
                  no_fidelius: bool = False,
                  no_py_inject: bool = False,
                  encoding: str = 'utf-8'):
        self._skip_resolve = no_resolve
        self._skip_env_loading = no_env_load
        self._skip_includes = no_includes
        self._skip_extends = no_extend
        self._skip_fidelius = no_fidelius
        self._skip_py_inject = no_py_inject

        self._file_name = os.path.basename(file_name)
        self._file_path = os.path.dirname(os.path.abspath(file_name))
        if not os.path.exists(file_name):
            raise AlvissFileNotFoundError('File not found', file_name=file_name)

        with open(file_name, 'r', encoding=encoding) as fin:
            self.load_raw(fin.read(), no_resolve=no_resolve, no_extend=no_extend,
                          no_includes=no_includes, no_env_load=no_env_load, no_fidelius=no_fidelius)

    def _extend(self):
        if self._extends:
            for inc_file, location in self._extends:
                extended = self.__class__()

                ext_file = os.path.join(self._file_path, inc_file)
                if not os.path.exists(ext_file):
                    raise AlvissFileNotFoundError('File to extend not found', file_name=ext_file)

                extended.load_file(ext_file, no_resolve=True, no_env_load=self._skip_env_loading, no_fidelius=True)

                old_data = self._data
                old_unresolved = self._unresolved
                new_data = {}
                new_unresolved = {}
                if location:
                    iters.nested_set(new_data, location, extended.data)
                else:
                    new_data = extended.data

                for old_loc, val in extended._unresolved.items():
                    # Only if we haven't overwritten this key yet!
                    new_key = location + old_loc if location else old_loc
                    if iters.nested_get(self._data, new_key) is None:
                        new_unresolved[location + old_loc] = val

                self._data = new_data
                iters.nested_dict_update(self._data, old_data)
                self._unresolved = new_unresolved
                self._unresolved.update(old_unresolved)

    def _includes(self):
        if self._include:
            for inc_file, location in self._include:
                inc_loader = self.__class__()
                full_file = os.path.join(self._file_path, inc_file)
                if not os.path.exists(full_file):
                    raise AlvissFileNotFoundError('File to include not found', file_name=full_file)

                inc_loader.load_file(full_file, no_resolve=True, no_env_load=self._skip_env_loading, no_fidelius=True)
                new_data = {}
                new_unresolved = {}
                if location:
                    iters.nested_set(new_data, location, inc_loader.data)
                    for old_loc, val in inc_loader._unresolved.items():
                        new_unresolved[location + old_loc] = val
                else:
                    new_data = inc_loader.data
                    new_unresolved = inc_loader._unresolved
                iters.nested_dict_update(self._data, new_data)
                self._unresolved.update(new_unresolved)

    def _resolve(self):
        while self._unresolved:
            start_count = self._resolve_count

            unresolved = list(self._unresolved.items())
            for location, value in unresolved:
                res = self._check_for_var(value, location)
                # if res:  # Not sure if this will cause some unwanted side effects?
                # (e.g. overwriting something with ''?)
                if res is not None:
                    iters.nested_set(self._data, location, res)

            if start_count == self._resolve_count:
                # Loop!
                break

        # Check for required vars...
        if self._unresolved:
            required_list = []
            unresolved = list(self._unresolved.items())
            for location, value in unresolved:
                for match in list(_VAR_PATTERN.finditer(value))[::-1]:
                    key = match.group(1)
                    if key.strip().endswith('!='):
                        required_list.append((location, value))
            if required_list:
                for location, value in required_list:
                    log.error(f'Required variable config value is unresolved: {location}, {value}')
                raise AlvissSyntaxError(f'Required variable config values were unresolved: {required_list!r}')

    def _fetch_fidelius(self):
        if self._fidelius_keys:
            try:
                from fidelius.fideliusapi import FideliusFactory
                from fidelius.fideliusapi import FideliusAppProps
            except ImportError:
                if self.get_fidelius_mode() == FideliusMode.ON_DEMAND:
                    for path_tuple, value in self._fidelius_keys.items():
                        log.error(f'Fidelius tag ({path_tuple}, {value}) found but Fidelius is not installed (ALVISS_FIDELIUS_MODE=ON_DEMAND)')
                raise AlvissFideliusNotInstalledError(f'Fidelius tags found but Fidelius is not installed (ALVISS_FIDELIUS_MODE=ON_DEMAND)')

            app_var = iters.nested_get(self.data, ('app', 'slug'), iters.nested_get(self.data, ('app', 'module_name')))
            if not app_var:
                raise AlvissFideliusSyntaxError('unable to resolve fidelius keys without app.slug or app.module_name')
            group_var = iters.nested_get(self.data, ('app', 'group'))
            if not group_var:
                raise AlvissFideliusSyntaxError('unable to resolve fidelius keys without app.group')
            env_var = iters.nested_get(self.data, ('app', 'env'))
            if not env_var:
                raise AlvissFideliusSyntaxError('unable to resolve fidelius keys without app.env')

            fidcls = FideliusFactory.get_class('mock' if self.get_fidelius_mode() == FideliusMode.MOCK else 'paramstore')

            ps = fidcls(FideliusAppProps(app=app_var,
                                         group=group_var,
                                         env=env_var),
                        **self._fidelius_kwargs or {})
            done_keys = []
            for path_tuple, value in self._fidelius_keys.items():
                new_value = ps.replace(value)
                if new_value != value:
                    iters.nested_set(self._data, path_tuple, new_value)
                    done_keys.append(path_tuple)

            if done_keys:
                for k in done_keys:
                    if k in self._fidelius_keys:
                        del self._fidelius_keys[k]
                    if k in self._unresolved:
                        del self._unresolved[k]

    def _load_value(self, v: Any, path: List[str]) -> Any:
        if isinstance(v, str):
            new_v = self._check_for_var(v, path)
            if new_v is not None:
                v = new_v

        elif isinstance(v, dict):
            v = self._load_dict(v, path)

        elif isinstance(v, list):
            v = self._load_list(v, path)

        return v

    def _load_list(self, lst: List, path: List[Union[str, int]]) -> List:
        new_l = []
        for i, o in enumerate(lst):
            path.append(i)
            new_l.append(self._load_value(o, path))
            path.pop()
        return new_l

    def _load_dict(self, d: Dict, path: Optional[List[str]] = None) -> Dict:
        path = path or []
        new_d = {}
        for k, v in d.items():
            path.append(k)
            if k in self._special_key_map:
                self._special_key_map[k](v, path)

            else:
                if k.startswith('__'):
                    path.pop()
                    continue
                elif '.' in k:
                    self._chains.append((tuple(path[:-1]+k.split('.')), self._load_value(v, path)))
                else:
                    new_d[k] = self._load_value(v, path)

            path.pop()
        return new_d

    def _set_chains(self):
        for location, value in self._chains:
            if len(location) > 1:
                key = location[-1]
                location = list(location[:-1])
                value = {key: value}

            old = iters.nested_get(self._data, location) or {}
            new_value = self._load_value(value, location) or {}
            iters.nested_dict_update(old, new_value)
            iters.nested_set(self._data, location, old)

    def _check_for_var(self, value: str, path: Sequence[str]) -> Optional[str]:
        path_key = tuple(self._explode_path_strings(path))

        unresolved = False
        updated = False
        for match in list(_VAR_PATTERN.finditer(value))[::-1]:
            res = self._try_to_resolve(match, path)
            if res is not None:
                before = value
                start = value[0:match.start()]
                end = value[match.end():]
                if not start and not end:
                    value = res
                else:
                    value = value[0:match.start()]+str(res)+value[match.end():]
                if value != before:
                    updated = True
            else:
                unresolved = True

        if unresolved:
            self._unresolved[path_key] = value
        else:
            if path_key in self._unresolved:
                del self._unresolved[path_key]

        if updated:
            if isinstance(value, (str, bytes)) and _VAR_PATTERN.search(value):
                # Still more left to resolve
                self._unresolved[path_key] = value
            return value

        return None

    @staticmethod
    def _explode_path_strings(path_strings: Union[Sequence[str], str]) -> List[str]:
        new_list = []
        if not isinstance(path_strings, (tuple, list)):
            path_strings: List[str] = [path_strings]

        for path in path_strings:
            if not isinstance(path, str):
                new_list += [path]
            else:
                new_list += path.split('.')

        return new_list

    @staticmethod
    def _get_env(key: str) -> str:
        if '=' in key:
            key, default = key.split('=', 2)
            if isinstance(key, str):
                if key.endswith('!'):
                    key = key[0:-1]
            return os.environ.get(key[8:], default)
        return os.environ.get(key[8:], None)

    def _try_to_resolve(self, match: re.Match, path: Sequence[str]) -> Optional[str]:
        if not self._data or self._skip_resolve:
            return None

        key = match.group(1)

        if not self._skip_env_loading and key.startswith('__ENV__:'):
            val = self._get_env(key)
            if val is not None:
                if val == '' and key.strip().endswith('!='):
                    return None  # Still unresolved
                self._resolve_count += 1
                return val
            return match.group(0)

        if not self._skip_fidelius and key.startswith('__FID__:'):
            if self.get_fidelius_mode() == FideliusMode.DISABLED:
                return match.group(0)

            elif self.get_fidelius_mode() == FideliusMode.SUBSTITUTE_ENV:
                subst_key = f'__ENV__:{key[8:].replace(":", "__")}'
                val = self._get_env(subst_key)
                if val is not None:
                    self._resolve_count += 1
                    return val
                return match.group(0)

            self._fidelius_keys[tuple(path)] = match.group(0)

        if not self._skip_py_inject and key.startswith('__PY__:'):
            val = self._py_inject(key)
            if val is not None:
                if val == '' and key.strip().endswith('!='):
                    return None  # Still unresolved
                self._resolve_count += 1
                return val
            return match.group(0)

        val = self._get_str_key(key)
        if val is not None:
            self._resolve_count += 1
        return val

    def _py_inject(self, key: str) -> Any:
        default = None
        if '=' in key:
            key, default = key.split('=', 2)
            if isinstance(key, str):
                if key.endswith('!'):
                    key = key[0:-1]

        return strimp.get_any(key[7:], default)

    def _get_str_key(self, key: str) -> Optional[str]:
        return iters.nested_get(self._data, key.split('.'))

    def _special_key_extends(self, value: Union[str, List[str]], path: List[str]):
        if not isinstance(value, list):
            value = [value]
        self._extends.extend([(v, tuple(path[:-1])) for v in value])

    def _special_key_include(self, value: Union[str, List[str]], path: List[str]):
        if not isinstance(value, list):
            value = [value]
        self._include.extend([(v, tuple(path[:-1])) for v in value])

    def _special_key_fidelius(self, value: Dict[str, Any], path: List[str]):
        """The `__fidelius__` special key should contain a dict (map) of
        fidelius configuration values which at the moment can be:

        - `ALVISS_FIDELIUS_MODE: [ON_DEMAND|ENABLED|DISABLED|SUBSTITUTE_ENV|MOCK]` which override the environment variable if any
        - `kwargs: [dict/map]` of key/values pairs which will be given to the fidelius __init__ call and can include (for the paramstore implementation):
            - `aws_access_key_id` - overrides the `FIDELIUS_AWS_ACCESS_KEY_ID`/`AWS_ACCESS_KEY_ID` otherwise taken from environment variables
            - `aws_secret_access_key` - overrides the `FIDELIUS_AWS_SECRET_ACCESS_KEY`/`AWS_SECRET_ACCESS_KEY` otherwise taken from environment variables
            - `aws_key_arn` - overrides the `FIDELIUS_AWS_KEY_ARN` otherwise taken from environment variable
            - `aws_region_name` - overrides the `FIDELIUS_AWS_REGION_NAME`/`AWS_DEFAULT_REGION` otherwise taken from environment variable
            - `aws_endpoint_url` - overrides the `FIDELIUS_AWS_ENDPOINT_URL` otherwise taken from environment variable (if any, it's optional)
            - `flush_cache_every_time` - set to true in order to flush fidelius' cache after every call (for testing purposes)

        :param value:
        :type value:
        :param path:
        :type path:
        :return:
        :rtype:
        """
        if not isinstance(value, dict):
            log.error("The '__fidelius__' special key must be a map/dict of key/pairs")
            return

        if 'ALVISS_FIDELIUS_MODE' in value:
            self.set_fidelius_mode(value['ALVISS_FIDELIUS_MODE'])
            log.info(f'Fidelius mode loaded from `__fidelius__` special key: {self.get_fidelius_mode().name}')
            if self.get_fidelius_mode() == FideliusMode.ENABLED:
                try:
                    import fidelius
                except ImportError:
                    log.error('ALVISS_FIDELIUS_MODE is ENABLED but fidelius is not installed')
                    raise AlvissFideliusNotInstalledError('ALVISS_FIDELIUS_MODE is ENABLED but fidelius is not installed')

        if 'kwargs' in value:
            if not isinstance(value['kwargs'], dict):
                log.error("The '__fidelius__.kwargs' special key must be a map/dict of key/pairs")
                return
            self._fidelius_kwargs: Dict = value['kwargs']
            disp = self._fidelius_kwargs.copy()
            if 'aws_secret_access_key' in disp:
                disp['aws_secret_access_key'] = '********************'

            log.info(f'Fidelius kwargs loaded from `__fidelius__` special key: {disp!r}')





__all__ = [
    'is_keyword',
    'escape_keyword',
    'unescape_keyword',
]


_KW_FIELDS = {  # These can't be field names in dataclasses
    'False',
    'None',
    'True',
    'and',
    'as',
    'assert',
    'async',
    'await',
    'break',
    'class',
    'continue',
    'def',
    'del',
    'elif',
    'else',
    'except',
    'finally',
    'for',
    'from',
    'global',
    'if',
    'import',
    'in',
    'is',
    'lambda',
    'nonlocal',
    'not',
    'or',
    'pass',
    'raise',
    'return',
    'try',
    'while',
    'with',
    'yield',
}


def is_keyword(string: str, include_escaped: bool = False) -> bool:
    """Given a string, returns True if it is a Python keyword (and thus can't
    be used as a field name in a dataclass).

    :param string: The string to check
    :param include_escaped: Also checks if the given string is an escaped
                            keyword (keyword with an appended underscore). False
                            by default.
    :return: True if the string is a Python keyword')
    """
    if include_escaped and string[-1] == '_':
        string = string[:-1]
    return string in _KW_FIELDS


def escape_keyword(string: str) -> str:
    """Returns an escaped version of the given string, if it is a Python
    keyword. Escaping just involves adding an appended underscore to its name.

    :param string: The string to escape
    :return: The string with an appended underscore, if it was a Python keyword,
             but otherwise the same string, unaltered.
    """
    if is_keyword(string):
        return f'{string}_'
    return string


def unescape_keyword(string: str) -> str:
    """If the given string is an escaped Python keyword (i.e. a keyword with an
    appended underscore) removes the underscore and returns that keyword.
    Otherwise the given string is returned unaltered.

    :param string: The string to unescape
    :return: The string with its underscore removed, if it was an escaped Python
             keyword
    """
    if string[-1] != '_':
        return string  # Doesn't need unescaping!
    check_string = string[:-1]
    if is_keyword(check_string):
        return check_string
    return string



__all__ = [
    "BaseConfig",
]
from ccptools.structs import *
from ccptools.tpu import iters
from ccptools.tpu import string
import json
import yaml
import collections
from alviss.utils import *


class _KwSafeEmptyDict(EmptyDict):
    def __getattribute__(self, name):
        name = unescape_keyword(name)
        if hasattr(collections.defaultdict, name):
            return collections.defaultdict.__getattribute__(self, name)
        if collections.defaultdict.__contains__(self, name):
            data = collections.defaultdict.__getitem__(self, name)
            if isinstance(data, dict):
                return _KwSafeEmptyDict(**data)
            elif data is None:
                return Empty
            else:
                return data
        return Empty

    def __getitem__(self, item):
        item = unescape_keyword(item)
        if collections.defaultdict.__contains__(self, item):
            data = collections.defaultdict.__getitem__(self, item)
            if isinstance(data, dict):
                return _KwSafeEmptyDict(**data)
            elif data is None:
                return Empty
            else:
                return data
        return Empty


class BaseConfig(object):
    """This object simplifies accessing configuration values by more-or-less
    behaving like a `ccptools.structs.EmptyDict`.

    That is to say, just access any value, no matter how nested it is, via dot
    attributes. If any part of the requested value wasn't in the loaded
    configuration file, the resulting value will just be an `Empty` (from
    `ccptools.structs.Empty`).

    Given the following `example.yaml` config file:
    ```yaml
    plugins:
        foo_plug:
            enabled: true
            foo_config: bar or something
    ```

    Loading this with Alviss as a BaseConfig object will result in the following:
    ```python
    >>> from alviss import quickloader
    >>> cfg = quickloader.autoload('example.yaml')
    >>> print(cfg.plugins.foo_plug.enabled)
    True
    >>> print(cfg.plugins.foo_plug.foo_config)
    bar or something
    >>> print(cfg.plugins.bar_plug.enabled)
    ```


    """
    _secret_keys = {'pass', 'secret', 'token', 'key'}

    def __init__(self, **kwargs):
        super().__setattr__('_data', _KwSafeEmptyDict(**kwargs))

    def __getattr__(self, item):
        return _KwSafeEmptyDict.__getattribute__(self._data, item)

    def __setattr__(self, key, value):
        if key.startswith('_'):
            super().__setattr__(key, value)
        self.update(**{key: value})

    def __str__(self):
        return _KwSafeEmptyDict.__str__(self._repr_dump(self._data))  # noqa

    def __repr__(self):
        return _KwSafeEmptyDict.__repr__(self._repr_dump(self._data))  # noqa

    def as_json(self, unmaksed: bool = False) -> str:
        return json.dumps(self.as_dict(unmaksed=unmaksed), indent=4)

    def as_yaml(self, unmaksed: bool = False) -> str:
        return yaml.dump(self.as_dict(unmaksed=unmaksed))

    def as_dict(self, unmaksed: bool = False) -> dict:
        return self._data if unmaksed else self._repr_dump(self._data)

    @classmethod
    def _repr_dump(cls, cfg_map: dict):
        d = {}
        for k, v in cfg_map.items():
            if isinstance(v, dict):
                d[k] = cls._repr_dump(v)
            elif isinstance(v, (str, bytes)) and cls._is_key_secret(k):
                d[k] = '********'
            else:
                d[k] = v
        return d

    @classmethod
    def _is_key_secret(cls, key: str):
        key = string.str_norm(key)
        for sk in cls._secret_keys:
            if sk in key:
                return True
        return False

    def load(self, **kwargs):
        super().__setattr__('_data', _KwSafeEmptyDict(**kwargs))

    def update(self, **kwargs):
        iters.nested_dict_update(self._data, _KwSafeEmptyDict(**kwargs))




__all__ = [
    'SimpleStaticRenderer',
]

from .interface import *
from alviss import quickloader
from alviss.structs.errors import *

import os
import pathlib

import logging
log = logging.getLogger(__file__)


class SimpleStaticRenderer(IStaticRenderer):
    def render_static_config_from_file(self, file: str) -> str:
        return quickloader.render_load(file)

    def render_static_config_to_file(self, input_file: str, output_file: str, overwrite_existing: bool = False):
        out = pathlib.Path(output_file).absolute()
        if out.exists() and not overwrite_existing:
            raise AlvissFileAlreadyExistsError('Output file already exists', file_name=output_file)

        results = self.render_static_config_from_file(input_file)

        if not out.parent.exists():
            log.debug(f'Creating output path: {out.parent}')
            os.makedirs(out.parent, exist_ok=True)

        with open(output_file, 'w') as fin:
            fin.write(results)

        return



import argparse

from alviss import __version__ as version
from alviss.renderers import SimpleStaticRenderer
from alviss.structs.errors import *

import sys


def main():
    parser = argparse.ArgumentParser(description='Renders a single static config from an Alviss formatted file and '
                                                 'its extends, includes and other expressions.',
                                     epilog=f'Alviss version {version}')

    parser.add_argument('file', help='The Alviss formatted config file to read, parse and render')
    parser.add_argument('-o', '--output', help='File to write the results to (otherwise its just printed to stdout)',
                        default='', nargs='?')
    parser.add_argument('-f', '--force-overwrite', help='Overwrite existing output file if it exists',
                        action='store_true')

    loudness_group = parser.add_mutually_exclusive_group()
    loudness_group.add_argument('-s', '--silent', action='store_true',
                                help='Only outputs the resulting rendered config (and errors print to stderr)')
    loudness_group.add_argument('-v', '--verbose', action="store_true",
                                help='Spits out DEBUG level logs')

    args = parser.parse_args()

    if args.verbose:
        import logging
        logging.basicConfig(level=logging.DEBUG)

    if not args.silent:
        print(f'Reading and parsing file: {args.file}...')

    try:
        if args.output:
            if not args.silent:
                print(f'Writing output to: {args.output}...')

            SimpleStaticRenderer().render_static_config_to_file(input_file=args.file,
                                                                output_file=args.output,
                                                                overwrite_existing=args.force_overwrite)

        else:
            if not args.silent:
                print(f'Printing results:')
                print(f'==================================================')
            print(SimpleStaticRenderer().render_static_config_from_file(args.file))

            if not args.silent:
                print(f'==================================================')

        if not args.silent:
            print(f'Done!')

    except AlvissError as e:
        if not args.silent:
            print(f'An error occurred: {e!r}')
        else:
            print(f'An error occurred: {e!r}', file=sys.stderr)
        sys.exit(3)


if __name__ == '__main__':
    main()



__all__ = [
    'IStubMaker',
]
from alviss.structs import *


class IStubMaker(abc.ABC):
    @abc.abstractmethod
    def render_stub_classes_from_descriptor_file(self, file: str, is_private: bool = True,
                                                 class_name: str = 'AlvissConfigStub') -> str:
        """Renders a Python module file with type hinting stub classes from the
        Alviss config type descriptor file.
        """
        pass

    @abc.abstractmethod
    def render_stub_classes_to_file(self, input_file: str, output_file: str,
                                    overwrite_existing: bool = False, is_private: bool = True,
                                    class_name: str = 'AlvissConfigStub'):
        """Writer the results of the `render_stub_classes_from_descriptor_file`
        call to the given output file.
        """
        pass



__all__ = [
    'SimpleStubMaker',
]
from alviss.quickloader import autoload
from .interface import *
from ._structs import *
from alviss.structs.errors import *

import os
import pathlib

import logging
log = logging.getLogger(__file__)


class SimpleStubMaker(IStubMaker):
    def render_stub_classes_from_descriptor_file(self, file, is_private: bool = True,
                                                 class_name: str = 'AlvissConfigStub') -> str:
        cfg = autoload(file)
        root_stub = StubClass.from_dict(cfg.as_dict(unmaksed=True), is_private=is_private)
        res = []
        class_names = []
        for stub in root_stub.get_all_sub_stubs():
            if not stub.class_name.startswith('_'):
                class_names.append(stub.class_name)
            res.append(stub.render_class_str())
        if not root_stub.class_name.startswith('_'):
            class_names.append(root_stub.class_name)
        res.append(root_stub.render_class_str())
        class_str = '\n\n\n'.join(res)

        if class_name:
            class_names.append(class_name)

        all_str = ''
        if class_names:
            all_str = '\n'.join([f"    '{c}'," for c in class_names])
            all_str = f"""__all__ = [
{all_str}
]

"""
        root_cls = ''
        if class_name:
            root_cls = f"""


class {class_name}(BaseConfig, {root_stub.class_name}):
    pass"""

        return f"""{all_str}from typing import *
from alviss.structs import Empty
from alviss.structs.cfgstub import _BaseCfgStub
from alviss.structs import BaseConfig


{class_str}{root_cls}"""

    def render_stub_classes_to_file(self, input_file: str, output_file: str, overwrite_existing: bool = False, is_private: bool = True, class_name: str = 'AlvissConfigStub'):
        out = pathlib.Path(output_file).absolute()
        if out.exists() and not overwrite_existing:
            raise AlvissFileAlreadyExistsError('Output file already exists', file_name=output_file)

        results = self.render_stub_classes_from_descriptor_file(input_file, is_private=is_private, class_name=class_name)

        if not out.parent.exists():
            log.debug(f'Creating output path: {out.parent}')
            os.makedirs(out.parent, exist_ok=True)

        with open(output_file, 'w') as fin:
            fin.write(results)
            fin.write('\n')

        return



__all__ = [
    'IAlvissLoader',
    'FideliusMode',
]

from alviss.structs import *


class FideliusMode(EnumEx):
    """Setting the environment variable ALVISS_FIDELIUS_MODE to any of these
    strings will override the default Fidelius functionality.

    E.g. this will just disable Fidelius:
    ```bash
    export ALVISS_FIDELIUS_MODE=DISABLED
    ```
    """
    ON_DEMAND = 0  # Load and use Fidelius if __FID__ tags are encountered (default)
    ENABLED = 1  # Load Fidelius on startup
    DISABLED = 2  # Fidelius is disabled and __FID__ tags are ignored
    SUBSTITUTE_ENV = 3  # Fidelius is disabled and __FID__ are substituted by __ENV__ (e.g. for testing or dev)
    MOCK = 4  # Fidelius is enabled but the Mock instance will be used (for testing and/or dev)


class IAlvissLoader(abc.ABC):
    @abc.abstractmethod
    def set_fidelius_mode(self, mode: Union[FideliusMode, str, int]):
        """Manually sets the operational mode of Fidelius. Otherwise it's
        determined by the `ALVISS_FIDELIUS_MODE` environment variable.
        """
        pass

    @abc.abstractmethod
    def get_fidelius_mode(self) -> FideliusMode:
        """Returns the current operational mode of Fidelius.
        """
        pass

    @abc.abstractmethod
    def load_file(self,
                  file_name: str,
                  no_resolve: bool = False,
                  no_extend: bool = False,
                  no_includes: bool = False,
                  no_env_load: bool = False,
                  no_fidelius: bool = False,
                  no_py_inject: bool = False,
                  encoding: str = 'utf-8'):
        """Loads a config file.

        :param file_name: The file to load.
        :param no_resolve: Skips resolving any special tags
        :param no_extend:
        :param no_includes:
        :param no_env_load:
        :param no_fidelius:
        :param no_py_inject:
        :param encoding:
        """
        pass

    # @abc.abstractmethod
    # def load_url(self, url: str):
    #     pass

    @abc.abstractmethod
    def load_raw(self, raw_data: str, no_resolve: bool = False,
                 no_extend: bool = False, no_includes: bool = False, no_env_load: bool = False,
                 no_fidelius: bool = False, no_py_inject: bool = False):
        """Loads configuration data directly from a string.

        The `load_file` method basically just reads a file and passes its
        content to this method in most cases."""
        pass

    @property
    @abc.abstractmethod
    def data(self) -> Dict:
        """A python dict with the raw configuration values loaded and parsed
        """
        pass

    @property
    @abc.abstractmethod
    def rendered(self) -> str:
        """A string of the configuration values loaded and parser in the same
        format as the initial input format
        """
        pass



__all__ = [
    'YamlFileConfigLoader',
]

from .base import *
import yaml

import logging
log = logging.getLogger(__name__)


class YamlFileConfigLoader(BaseLoader):
    def load_raw(self, raw_data: str, no_resolve: bool = False,
                 no_extend: bool = False, no_includes: bool = False, no_env_load: bool = False,
                 no_fidelius: bool = False, no_py_inject: bool = False):
        self._skip_resolve = no_resolve
        self._skip_env_loading = no_env_load
        self._skip_includes = no_includes
        self._skip_extends = no_extend
        self._skip_fidelius = no_fidelius
        self._skip_py_inject = no_py_inject
        self._data = self._load_dict(yaml.safe_load(raw_data))
        self._set_chains()
        if not no_extend:
            self._extend()
        if not no_includes:
            self._includes()
        if not no_resolve:
            self._resolve()
        if not no_fidelius:
            self._fetch_fidelius()

    @property
    def rendered(self) -> str:
        return yaml.dump(self._data)



__all__ = [
    'StubField',
    'StubClass',
]
import dataclasses
from typing import *
from alviss.structs.errors import *
from alviss.utils import *
import re

import logging
log = logging.getLogger(__name__)


_WORD_PATTERN = re.compile(r'(\w+)')

_VALID_TYPE_NAME_SET = {
    'str',
    'int',
    'float',
    'bool',

    'None',
    'Empty',
    'Dict',
    'List',
    'Any',
    'Union'
}

_CAP_WORD_SET = {
    'none',
    'empty',
    'dict',
    'list',
    'any',
    'union',
}


@dataclasses.dataclass
class StubField:
    name: str
    value: Union[str, List[str], 'StubClass']
    is_self_required: bool
    ancestors: List[str] = dataclasses.field(default_factory=list)
    is_dict_list: bool = False
    is_map_of_stuff: bool = False

    def __post_init__(self):
        self._fix_and_validate_type_names()

    def _fix_type_name(self, type_name: str, has_bracket: bool = False) -> str:
        og_type_name = type_name
        type_name = type_name.strip().lower()
        if type_name in _CAP_WORD_SET:
            type_name = type_name.capitalize()
        if type_name not in _VALID_TYPE_NAME_SET:
            raise AlvissStubberInvalidTypeName(f'Invalid type name', field_name=self.name, type_name=og_type_name)

        if type_name == 'Dict' and not has_bracket:
            type_name = 'Dict[str,Any]'  # Default
        elif type_name == 'List' and not has_bracket:
            type_name = 'List[Any]'
        return type_name

    def _loop_over_type_words(self, string: str) -> str:
        new_s = []
        i = 0
        string = string.replace(' ', '')
        for match in _WORD_PATTERN.finditer(string):
            if match.start() > i:
                new_s.append(string[i:match.start()])
            word = string[match.start():match.end()]
            has_bracket = string[match.end():match.end()+1] == '['
            new_s.append(self._fix_type_name(word, has_bracket))
            i = match.end()
        if len(string) >= i:
            new_s.append(string[i:])
        return ''.join(new_s).replace(',', ', ')

    def _fix_and_validate_type_names(self):
        if isinstance(self.value, StubClass):
            return
        elif isinstance(self.value, list):
            new_list = []
            for v in self.value:
                new_list.append(self._loop_over_type_words(v))
            self.value = new_list
        else:
            self.value = self._loop_over_type_words(self.value)

    @property
    def safe_field_name(self) -> str:
        return escape_keyword(self.name)

    @property
    def is_required(self) -> bool:
        return self.is_self_required or self.has_required_children

    @property
    def has_required_children(self) -> bool:
        if isinstance(self.value, StubClass):
            return self.value.has_required_fields
        return False

    @classmethod
    def from_keyval(cls, key: str, value: Union[str, List, Dict], ancestors: Optional[List[str]],
                    is_map_of_stuff: bool = False,
                    pre_required: bool = False,
                    is_private: bool = True) -> 'StubField':
        required = pre_required

        if key.endswith('*'):
            required = True
            key = key[:-1].strip()

        if isinstance(value, dict):
            if len(value) == 1 and ('${str}' in value or '${str}*' in value):
                sub_key, sub_val = value.popitem()
                if sub_key.endswith('*'):
                    required = True
                return cls.from_keyval(key=key, value=sub_val, pre_required=required,
                                       ancestors=ancestors, is_map_of_stuff=True, is_private=is_private)

            else:
                value = StubClass.from_dict(input_dict_or_list=value, field_name=key, ancestors=ancestors, is_private=is_private)
                return cls(name=key, value=value, is_self_required=required,
                           ancestors=ancestors, is_map_of_stuff=is_map_of_stuff)

        elif isinstance(value, list):
            if len(value) != 1:
                raise AlvissStubberSyntaxError('Lists in type descriptor files must have exactly one element',
                                               field_name='.'.join(ancestors+[key]))

            if isinstance(value[0], dict):
                value = StubClass.from_dict(input_dict_or_list=value[0], field_name=key, ancestors=ancestors, is_private=is_private)
                return cls(name=key, value=value, is_self_required=required, ancestors=ancestors,
                           is_dict_list=True, is_map_of_stuff=is_map_of_stuff)

            elif isinstance(value[0], str):
                value = value[0].strip()

                if value.endswith('*'):
                    required = True
                    value = value[:-1]

                if '|' in value:  # Todo: this does not handle brackets, e.g. List[str|int]
                    value_list = [p.strip() for p in value.split('|')]
                    value = f'Union[{",".join(value_list)}]'
                value = f'List[{value}]'

                # Then continue below as if it's a string!

            else:
                raise AlvissStubberSyntaxError(f'Unexpected type in list of descriptor file: {type(value[0])}',
                                               field_name='.'.join(ancestors+[key]))
        elif isinstance(value, str):
            if value.endswith('*'):
                required = True
                value = value[:-1].strip()
            if '|' in value:  # Todo: this does not handle brackets, e.g. List[str|int]
                value = [p.strip() for p in value.split('|')]

        return cls(name=key, value=value, is_self_required=required, ancestors=ancestors.copy(),
                   is_map_of_stuff=is_map_of_stuff)

    def render_field_str(self) -> str:
        if isinstance(self.value, StubClass):
            if self.is_required:
                if self.is_dict_list:
                    return f'    {self.safe_field_name}: List[{self.value.class_name}]'
                elif self.is_map_of_stuff:
                    return f'    {self.safe_field_name}: Dict[str, {self.value.class_name}]'
                else:
                    return f'    {self.safe_field_name}: {self.value.class_name}'
            else:
                if self.is_dict_list:
                    type_list = [f'List[{self.value.class_name}]', 'Empty']
                elif self.is_map_of_stuff:
                    type_list = [f'Dict[str, {self.value.class_name}]', 'Empty']
                else:
                    type_list = [self.value.class_name, 'Empty']

        elif isinstance(self.value, list):
            type_list = self.value
            if not self.is_required:
                type_list.append('Empty')
        else:
            if self.is_required:
                return f'    {self.safe_field_name}: {self.value}'
            else:
                type_list = [self.value, 'Empty']

        return f'    {self.safe_field_name}: Union[{", ".join(type_list)}]'


@dataclasses.dataclass
class StubClass:
    name: str
    ancestors: List[str] = dataclasses.field(default_factory=list)
    fields: List[StubField] = dataclasses.field(default_factory=list)
    is_private: bool = True

    @property
    def has_required_fields(self) -> bool:
        for field in self.fields:
            if field.is_required:
                return True
        return False

    @classmethod
    def from_dict(cls,
                  input_dict_or_list: Union[Dict[str, Any], List[Any]],
                  field_name: str = '',
                  ancestors: Optional[List[str]] = None,
                  is_private: bool = True) -> 'StubClass':
        ancestors = ancestors or []
        ancestors.append(field_name)
        if isinstance(input_dict_or_list, dict):
            fields = [StubField.from_keyval(k, v, ancestors.copy(), is_private=is_private) for k, v in input_dict_or_list.items()]
        elif isinstance(input_dict_or_list, list):  # Should you ever get a list...?!?
            log.warning('NOT SURE THIS SHOULD EVER HAPPEN!!!')
            fields = [StubField.from_keyval('__list__', input_dict_or_list[0], ancestors.copy(), is_private=is_private)]
        else:
            raise AlvissStubberSyntaxError(f'Unexpected type fed to StubClass.from_dict: {type(input_dict_or_list)}')
        ancestors.pop()
        return cls(name=field_name, ancestors=ancestors, fields=fields, is_private=is_private)

    @staticmethod
    def field_name_to_class_name(field_name: str) -> str:
        if not field_name.isupper() and not field_name.islower():  # Mixed Casing
            tmp = []
            for c in field_name:
                if c.isupper():
                    if not tmp:
                        tmp.append(c)
                    else:
                        tmp.append('_')
                        tmp.append(c)
                else:
                    tmp.append(c)
            field_name = ''.join(tmp)
        parts = field_name.split('_')
        buff = []
        for part in parts:
            if not part:
                continue
            buff.append(part.capitalize())
        result = ''.join(buff)
        return escape_keyword(result)  # Just in case!

    @property
    def class_name(self) -> str:
        start = 'Cfg'
        if self.is_private:
            start = '_Cfg'
        if not self.name:
            return f'{start}Stub'
        else:
            if self.ancestors:
                return f'{start}{"".join([self.field_name_to_class_name(a) for a in self.ancestors+[self.name]])}Stub'
            else:
                return f'{start}{self.field_name_to_class_name(self.name)}Stub'

    def render_class_str(self) -> str:
        lines = [f'class {self.class_name}(_BaseCfgStub, dict):']
        for field in self.fields:
            lines.append(field.render_field_str())
        return '\n'.join(lines)

    def get_all_sub_stubs(self) -> List['StubClass']:
        lst = []
        for field in self.fields:
            if isinstance(field.value, StubClass):
                lst.extend(field.value.get_all_sub_stubs())
                lst.append(field.value)
        return lst


if __name__ == '__main__':
    d = StubClass.from_dict({
        'foo': 'int*',
        'bar': 'str|int',
        'sub_shit': {
            'extra': {
                'foo': 'str',
                'bar': 'str|bool*',
            },
            'bar': 'float',
        }
    })
    for s in d.get_all_sub_stubs():
        print(s.render_class_str())
        print('----------')
    print(d.render_class_str())


__all__ = [
    'IStaticRenderer',
]
import abc


class IStaticRenderer(abc.ABC):
    @abc.abstractmethod
    def render_static_config_from_file(self, file: str) -> str:
        """Renders a single static configuration file from an Alviss formatted
        file, including all included and/or extended files and resolving all
        expressions, variables and internal references and such.
        """
        pass

    @abc.abstractmethod
    def render_static_config_to_file(self, input_file: str, output_file: str, overwrite_existing: bool = False):
        """Writer the results of the `render_static_config_from_file` call to
        the given output file.
        """
        pass



__all__ = [
    'autoload',
    'guess_loader_class',
    'raw_load',
    'render_load',
]

import os

from alviss.structs import *
from .interface import *
from .jsonfile import *
from .yamlfile import *

_EXTENSION_LOADER_MAP = {
    '.json': JsonFileConfigLoader,
    '.yaml': YamlFileConfigLoader,
    '.yml': YamlFileConfigLoader,
}


def guess_loader_class(file_name: str) -> Type[IAlvissLoader]:
    """Returns the most probable loader class based of the given file's extension.
    """
    ext = os.path.splitext(file_name)[-1]
    loader = _EXTENSION_LOADER_MAP.get(ext, None)
    if not loader:
        raise AlvissUnknownFileTypeError(f'Dont know how to autoload file extension {ext}', file_name=file_name)
    return loader


def raw_load(file_name: str, skip_env_loading: bool = False, skip_fidelius: bool = False, skip_py_inject: bool = False, encoding: str = 'utf-8') -> Dict:
    """Loads and parses the given config file and returns the "raw" results as a
    simple python dictionary.

    This is mostly intended for the testing and debugging of the configuration
    files themselves if needed.

    :param file_name: The file to load
    :param skip_env_loading: Set to True to skip resolving environment variable tags
    :param skip_fidelius: Set to True to skip resolving Fidelius tags
    :param skip_py_inject: Set to True to skip injecting Python imported values
    :param encoding: Encoding to use when reading the file (utf-8 by default)
    :return: A dict with the results
    """
    loader = guess_loader_class(file_name)()
    loader.load_file(file_name, no_env_load=skip_env_loading, no_fidelius=skip_fidelius,
                     no_py_inject=skip_py_inject, encoding=encoding)
    return loader.data


def render_load(file_name: str, skip_env_loading: bool = True, skip_fidelius: bool = True, skip_py_inject: bool = True, encoding: str = 'utf-8') -> str:
    """Loads and parses the given config file and returns the "render" results
    as a str in the same format the original file used.

    The main point of this method is to be able to load and parse an Alviss
    configuration file, including all includes, extensions, internal references,
    environment variables, Fidelius values etc. and render as a single
    value/file of the same format.

    As a use-case example, it was originally built in order to generate
    Kubernetes YAML manifest files using Alviss' format to simplify the process
    (i.e. includes, internal references, secret injection via Fidelius during
    deployment and so on).

    :param file_name: The file to load
    :param skip_env_loading: Set to True to skip resolving environment variable
                             tags
    :param skip_fidelius: Set to True to skip resolving Fidelius tags
    :param skip_py_inject: Set to True to skip injecting Python imported values
    :param encoding: Encoding to use when reading the file (utf-8 by default)
    :return: A string representation of the configuration data loaded, in the
             same format as the target file if possible
    """
    loader = guess_loader_class(file_name)()
    loader.load_file(file_name, no_env_load=skip_env_loading, no_fidelius=skip_fidelius, no_py_inject=skip_py_inject, encoding=encoding)
    return loader.rendered


def autoload(file_name: str, encoding: str = 'utf-8') -> BaseConfig:
    """Loads and parses the given config file and returns as a `BaseConfig`
    object.

    :param file_name:
    :type file_name:
    :param encoding:
    :type encoding:
    :return:
    :rtype:
    """
    loader = guess_loader_class(file_name)()
    loader.load_file(file_name, encoding=encoding)
    return BaseConfig(**raw_load(file_name))



__all__ = [
    'AlvissError',
    'AlvissFileNotFoundError',
    'AlvissFileAlreadyExistsError',
    'AlvissParsingError',

    'AlvissSyntaxError',
    'AlvissExpressionResolvingError',
    'AlvissFideliusError',
    'AlvissFideliusNotInstalledError',
    'AlvissFideliusSyntaxError',

    'AlvissUnknownFileTypeError',
    'AlvissStubberError',
    'AlvissStubberSyntaxError',
    'AlvissStubberInvalidTypeName',
]


class AlvissError(Exception):
    pass


class AlvissFileNotFoundError(AlvissError, FileNotFoundError):
    def __init__(self, message: str, file_name: str = '?'):
        super().__init__(message)
        self.file_name = file_name

    def __str__(self):
        return f'{super().__str__()} (file_name="{self.file_name}")'


class AlvissFileAlreadyExistsError(AlvissError, FileExistsError):
    def __init__(self, message: str, file_name: str = '?'):
        super().__init__(message)
        self.file_name = file_name

    def __str__(self):
        return f'{super().__str__()} (file_name="{self.file_name}")'


class AlvissParsingError(AlvissError, ValueError):
    pass


class AlvissSyntaxError(AlvissParsingError):
    pass


class AlvissExpressionResolvingError(AlvissParsingError):
    pass


class AlvissFideliusError(AlvissExpressionResolvingError):
    pass


class AlvissFideliusNotInstalledError(AlvissExpressionResolvingError, ImportError):
    pass


class AlvissFideliusSyntaxError(AlvissExpressionResolvingError, AlvissSyntaxError):
    pass


class AlvissUnknownFileTypeError(AlvissParsingError, NotImplementedError):
    def __init__(self, message: str, file_name: str = '?'):
        super().__init__(message)
        self.file_name = file_name

    def __str__(self):
        return f'{super().__str__()} (file_name="{self.file_name}")'


class AlvissStubberError(AlvissError):
    pass


class AlvissStubberSyntaxError(AlvissStubberError, ValueError):
    def __init__(self, message: str, field_name: str = '?'):
        super().__init__(message)
        self.field_name = field_name

    def __str__(self):
        return f'{super().__str__()} (field_name="{self.field_name}")'


class AlvissStubberInvalidTypeName(AlvissStubberSyntaxError, TypeError):
    def __init__(self, message: str, field_name: str = '?', type_name: str = '?'):
        super().__init__(message, field_name)
        self.type_name = type_name

    def __str__(self):
        return f'{super().__str__()} (type_name="{self.type_name}", field_name="{self.field_name}")'



import argparse

from alviss import __version__ as version
from alviss import stubber
from alviss.structs.errors import *
import sys


def main():
    parser = argparse.ArgumentParser(description='Generates Python dataclass stubs based on the given alviss type descriptor file.',
                                     epilog=f'Alviss version {version}')

    parser.add_argument('file', help='The Alviss config type descriptor file to generate strubs from.')
    parser.add_argument('-o', '--output', help='File to write the generated stub code to (otherwise its just printed to stdout)',
                        default='', nargs='?')
    parser.add_argument('-f', '--force-overwrite', help='Overwrite existing output file if it exists', action='store_true')
    parser.add_argument('-n', '--class-name',
                        help='The name of the resulting "final" stub class generated when outputting to a file (default is "AlvissConfigStub"). Set to "None" to skip generating the "final" class.',
                        default='AlvissConfigStub')
    parser.add_argument('-x', '--export-all', help='Make all stub class names public and export via __all__ when outputting to a file.',
                        action='store_true')

    loudness_group = parser.add_mutually_exclusive_group()
    loudness_group.add_argument('-s', '--silent', action='store_true',
                                help='Only outputs the resulting rendered code (and errors print to stderr)')
    loudness_group.add_argument('-v', '--verbose', action="store_true",
                                help='Spits out DEBUG level logs')

    args = parser.parse_args()

    if args.verbose:
        import logging
        logging.basicConfig(level=logging.DEBUG)

    if not args.silent:
        print(f'Reading and stubbing file: {args.file}...')

    cls_name = 'AlvissConfigStub' if args.class_name is None else args.class_name
    if cls_name.lower().strip() == 'none':
        cls_name = ''

    try:
        if args.output:
            if not args.silent:
                print(f'Writing output to: {args.output}...')

            stubber.SimpleStubMaker().render_stub_classes_to_file(input_file=args.file,
                                                                  output_file=args.output,
                                                                  overwrite_existing=args.force_overwrite,
                                                                  is_private=not args.export_all,
                                                                  class_name=cls_name)
        else:
            if not args.silent:
                print(f'Printing results:')
                print(f'==================================================')

            print(stubber.SimpleStubMaker().render_stub_classes_from_descriptor_file(args.file, class_name=cls_name, is_private=not args.export_all))

            if not args.silent:
                print(f'==================================================')

        if not args.silent:
            print(f'Done!')

    except AlvissError as e:
        if not args.silent:
            print(f'An error occurred: {e!r}')
        else:
            print(f'An error occurred: {e!r}', file=sys.stderr)
        sys.exit(3)


if __name__ == '__main__':
    main()



__all__ = [
    'JsonFileConfigLoader',
]

from .base import *
import json

import logging
log = logging.getLogger(__name__)


class JsonFileConfigLoader(BaseLoader):
    @staticmethod
    def load(file_name: str) -> dict:
        """
        LEGACY STUFF!
        """
        ldr = JsonFileConfigLoader()
        ldr.load_file(file_name)
        return ldr.data

    # def load_url(self, url: str):
    #     pass

    def load_raw(self, raw_data: str, no_resolve: bool = False,
                 no_extend: bool = False, no_includes: bool = False, no_env_load: bool = False,
                 no_fidelius: bool = False, no_py_inject: bool = False):
        self._skip_resolve = no_resolve
        self._skip_env_loading = no_env_load
        self._skip_includes = no_includes
        self._skip_extends = no_extend
        self._skip_fidelius = no_fidelius
        self._skip_py_inject = no_py_inject

        self._data = self._load_dict(json.loads(raw_data))
        self._set_chains()
        if not no_extend:
            self._extend()
        if not no_includes:
            self._includes()
        if not no_resolve:
            self._resolve()
        if not no_fidelius:
            self._fetch_fidelius()

    @property
    def rendered(self) -> str:
        return json.dumps(self._data, indent=4)



__all__ = [
    '_BaseCfgStub',
]
from ._base import *


class _BaseCfgStub(Protocol):
    """This is the base Protocol for all generated Stubs

    """
    def __getattr__(self, item) -> Union[Any, Empty]:
        ...
